//
// --------------- mesh warping by DiGiTaLFX 2011 ---------------
//

autowatch = 1;
inlets = 1;
outlets = 1;

// misc variables
var windowsize = [320, 240];
var mousebutton = 0;
var clickedvertex = null;

// control matrix
var controlMatrix = new JitterMatrix(3, "float32", 2, 2);
controlMatrix.setcell(0, 0, "val", -1, -0.8, 0);
controlMatrix.setcell(0, 1, "val", -1, 0.8, 0);
controlMatrix.setcell(1, 0, "val", 1, -0.8, 0);
controlMatrix.setcell(1, 1, "val", 1, 0.8, 0);

// interpolated vertex matrix
var vertexMatrix = new JitterMatrix(3, "float32", 20, 20);
vertexMatrix.interp = 1;

// texture matrix and options
var textureMatrix = new JitterMatrix(3, "float32", 20, 20);
var calcTexMap = false;

// sketch for additional drawing
var sketchpoints = true;
var sketch = new JitterObject("jit.gl.sketch", "map");
sketch.transform_reset = 2;
sketch.layer = 10;

post("mesh.js loaded");



//
// --------------- general settings ---------------
//

function loadbang() {

	outlet(0, "transform_reset", 2);
	output();

}

function dim(x, y) {

	vertexMatrix.dim = [x, y];
	textureMatrix.dim = [x, y];
	
	output();

}

function winsize(x, y) {

	windowsize = [x, y];

}



//
// --------------- mesh manipulation ---------------
//

function topleft(x, y, z) {

	controlMatrix.setcell(0, 1, "val", x, y, z);
	output();

}

function topright(x, y, z) {

	controlMatrix.setcell(1, 1, "val", x, y, z);
	output();

}

function bottomleft(x, y, z) {

	controlMatrix.setcell(0, 0, "val", x, y, z);
	output();

}

function bottomright(x, y, z) {

	controlMatrix.setcell(1, 0, "val", x, y, z);
	output();

}


function output() {
	
	vertexMatrix.frommatrix(controlMatrix);
	
	updateTexture();
	
	outlet(0, "jit_matrix", vertexMatrix.name);
	outlet(0, "texcoord_matrix", "jit_matrix", textureMatrix.name);
	
	if (sketchpoints) {drawpoints();}

}



//
// --------------- allows mouse to drag verticies ---------------
//

function mousemap(a, b, button) {

	if (mousebutton === 0) {
	
		if (button === 1) {
			
			for (i = 0; i < controlMatrix.dim[0]; i++) {
				for (j = 0; j < controlMatrix.dim[1]; j++) {
					if (arePointsNear(controlMatrix.getcell(i, j)[0], controlMatrix.getcell(i, j)[1], a, b)) {clickedvertex = [i, j];}
				}
			}

		}
	
	}
	
	if (mousebutton === 1) {
	
		if (button === 1) {
		
			if (clickedvertex !== null) {controlMatrix.setcell(clickedvertex, "val", a, b, 0);}
		
		}
		
		if (button === 0) {
		
			clickedvertex = null;
		
		}
	
	}
	
	output();
	
	mousebutton = button;

}

function arePointsNear(x1, y1, x2, y2) {

	return ( ((x1 > x2 - 0.1) && (x1 < x2 + 0.1)) && ((y1 > y2 - 0.1) && (y1 < y2 + 0.1)) );

}



//
// --------------- texture coordinate calculations ---------------
//

function customTexMap(enabled) {

	calcTexMap = enabled;

}
function updateTexture()
{

	var i, j;
	var A = [controlMatrix.getcell(0, 1)[0], controlMatrix.getcell(0, 1)[1]];
	var B = [controlMatrix.getcell(1, 1)[0], controlMatrix.getcell(1, 1)[1]];
	var C = [controlMatrix.getcell(1, 0)[0], controlMatrix.getcell(1, 0)[1]];
	var D = [controlMatrix.getcell(0, 0)[0], controlMatrix.getcell(0, 0)[1]];
	var x, y, z, w, O, f1, f2;
	
	if (!calcTexMap) {
		for (i = 0; i < textureMatrix.dim[0]; i++) {
		
			for (j = 0; j < textureMatrix.dim[1]; j++) {
			
				textureMatrix.setcell(i, j, "val", i*(1/(textureMatrix.dim[0] - 1)), j*(1/(textureMatrix.dim[1] - 1)), 0);
			
			}
		
		}
	}
	
	if (calcTexMap) {
	
		O = intersection(A, C, B, D);
		f1 = intersection(A, B, C, D);
		f2 = intersection(A, D, B, C);
		x = intersection(f1, O, B, C);
		y = intersection(f2, O, D, C);
		z = intersection(f1, O, A, D);
		w = intersection(f2, O, A, B);
		
		wp = [linelength(w, B) / (linelength(A, w) + linelength(w, B)), 1];
		yp = [linelength(y, C) / (linelength(D, y) + linelength(y, C)), 0];
		zp = [0, linelength(A, z) / (linelength(A, z) + linelength(z, D))];
		xp = [1, linelength(B, x) / (linelength(B, x) + linelength(x, C))];
		Op = [linelength(O, x) / (linelength(z, O) + linelength(O, x)), linelength(w, O) / (linelength(w, O) + linelength(O, y))];
		
		var temp = new JitterMatrix(3, "float32", 3, 3);
		
		temp.setcell(0, 2, "val", 0, 1, 0);
		temp.setcell(1, 2, "val", wp, 0);
		temp.setcell(2, 2, "val", 1, 1, 0);
		
		temp.setcell(0, 1, "val", zp, 0);
		temp.setcell(1, 1, "val", Op, 0);
		temp.setcell(2, 1, "val", xp, 0);
		
		temp.setcell(0, 0, "val", 0, 0, 0);
		temp.setcell(1, 0, "val", yp, 0);
		temp.setcell(2, 0, "val", 1, 0, 0);
		
		textureMatrix.interp = 1;
		textureMatrix.frommatrix(temp);
	
	}

}

function intersection(p1, p2, p3, p4) {

	x = ( (p1[0] * p2[1] - p1[1] * p2[0]) * (p3[0] - p4[0]) - (p1[0] - p2[0]) * (p3[0] * p4[1] - p3[1] * p4[0]) ) / ( (p1[0] - p2[0]) * (p3[1] - p4[1]) - (p1[1] - p2[1]) * (p3[0] - p4[0]) )
	y = ( (p1[0] * p2[1] - p1[1] * p2[0]) * (p3[1] - p4[1]) - (p1[1] - p2[1]) * (p3[0] * p4[1] - p3[1] * p4[0]) ) / ( (p1[0] - p2[0]) * (p3[1] - p4[1]) - (p1[1] - p2[1]) * (p3[0] - p4[0]) )
	
	return [x, y];

}

function linelength(p1, p2) {

	return Math.sqrt(Math.pow(p2[0] - p1[0], 2) * windowsize[0] + Math.pow(p2[1] - p1[1], 2) * windowsize[1])
	// not sure if window size actually matters in the calculation.

}



//
// --------------- shows handles to move corners ---------------
//

function showpoints(enabled) {

	if (enabled) {sketchpoints = true; drawpoints();}
	if (!enabled) {sketchpoints = false; sketch.reset();}

}

function drawpoints() {

	sketch.reset();
	sketch.glcolor(1, 0, 0, 0.7);
	
	for (i = 0; i < controlMatrix.dim[0]; i++) {
	
		for (j = 0; j < controlMatrix.dim[1]; j++) {
			
			sketch.moveto(controlMatrix.getcell(i, j)[0], controlMatrix.getcell(i, j)[1], 0.);
			sketch.framecircle(0.025);
			
		}
	
	}

}